# -*- coding: utf-8 -*-
"""Fake (in-memory only) attribute container store for testing."""

import ast
import collections
import copy

from acstore import interface
from acstore.containers import interface as containers_interface


class FakeAttributeContainerStore(interface.AttributeContainerStore):
  """Fake (in-memory only) attribute container store."""

  def __init__(self):
    """Initializes a fake (in-memory only) store."""
    super(FakeAttributeContainerStore, self).__init__()
    self._attribute_container_indexes = {}
    self._attribute_containers = {}
    self._is_open = False

  def _RaiseIfNotReadable(self):
    """Raises if the store is not readable.

     Raises:
       OSError: if the store cannot be read from.
       IOError: if the store cannot be read from.
    """
    if not self._is_open:
      raise IOError('Unable to read from closed storage writer.')

  def _RaiseIfNotWritable(self):
    """Raises if the storage file is not writable.

    Raises:
      IOError: when the storage writer is closed.
      OSError: when the storage writer is closed.
    """
    if not self._is_open:
      raise IOError('Unable to write to closed storage writer.')

  def _WriteExistingAttributeContainer(self, container):
    """Writes an existing attribute container to the store.

    Args:
      container (AttributeContainer): attribute container.

    Raises:
      IOError: if an unsupported identifier is provided or if the attribute
          container does not exist.
      OSError: if an unsupported identifier is provided or if the attribute
          container does not exist.
    """
    identifier = container.GetIdentifier()
    lookup_key = identifier.CopyToString()

    containers = self._attribute_containers.get(container.CONTAINER_TYPE, None)
    if containers is None or lookup_key not in containers:
      raise IOError((
          f'Missing attribute container: {container.CONTAINER_TYPE:s} with '
          f'identifier: {lookup_key:s}'))

    containers[lookup_key] = container

  def _WriteNewAttributeContainer(self, container):
    """Writes a new attribute container to the store.

    Args:
      container (AttributeContainer): attribute container.
    """
    containers = self._attribute_containers.get(container.CONTAINER_TYPE, None)
    if containers is None:
      containers = collections.OrderedDict()
      self._attribute_containers[container.CONTAINER_TYPE] = containers

    container_indexes = self._attribute_container_indexes.get(
        container.CONTAINER_TYPE, None)
    if container_indexes is None:
      container_indexes = []
      self._attribute_container_indexes[container.CONTAINER_TYPE] = (
          container_indexes)

    next_sequence_number = self._GetAttributeContainerNextSequenceNumber(
        container.CONTAINER_TYPE)

    identifier = containers_interface.AttributeContainerIdentifier(
        name=container.CONTAINER_TYPE, sequence_number=next_sequence_number)
    container.SetIdentifier(identifier)

    lookup_key = identifier.CopyToString()

    # Make sure the fake storage preserves the state of the attribute container.
    containers[lookup_key] = copy.deepcopy(container)
    container_indexes.append(lookup_key)

  def Close(self):
    """Closes the store.

    Raises:
      IOError: if the store is already closed.
      OSError: if the store is already closed.
    """
    if not self._is_open:
      raise IOError('Store already closed.')

    self._is_open = False

  def GetAttributeContainerByIdentifier(self, container_type, identifier):
    """Retrieves a specific type of container with a specific identifier.

    Args:
      container_type (str): container type.
      identifier (AttributeContainerIdentifier): attribute container identifier.

    Returns:
      AttributeContainer: attribute container or None if not available.
    """
    containers = self._attribute_containers.get(container_type, {})

    lookup_key = identifier.CopyToString()
    return containers.get(lookup_key, None)

  def GetAttributeContainerByIndex(self, container_type, index):
    """Retrieves a specific attribute container.

    Args:
      container_type (str): attribute container type.
      index (int): attribute container index.

    Returns:
      AttributeContainer: attribute container or None if not available.
    """
    containers = self._attribute_containers.get(container_type, {})
    number_of_containers = len(containers)
    if index < 0 or index >= number_of_containers:
      return None

    container_indexes = self._attribute_container_indexes.get(
        container_type, None)
    lookup_key = container_indexes[index]

    return containers[lookup_key]

  def GetAttributeContainers(self, container_type, filter_expression=None):
    """Retrieves a specific type of attribute containers.

    Args:
      container_type (str): attribute container type.
      filter_expression (Optional[str]): expression to filter the resulting
          attribute containers by.

    Yield:
      AttributeContainer: attribute container.
    """
    if filter_expression:
      expression_ast = ast.parse(filter_expression, mode='eval')
      filter_expression = compile(expression_ast, '<string>', mode='eval')

    for attribute_container in self._attribute_containers.get(
        container_type, {}).values():
      if attribute_container.MatchesExpression(filter_expression):
        yield attribute_container

  def GetNumberOfAttributeContainers(self, container_type):
    """Retrieves the number of a specific type of attribute containers.

    Args:
      container_type (str): attribute container type.

    Returns:
      int: the number of containers of a specified type.
    """
    containers = self._attribute_containers.get(container_type, {})
    return len(containers)

  def HasAttributeContainers(self, container_type):
    """Determines if a store contains a specific type of attribute container.

    Args:
      container_type (str): attribute container type.

    Returns:
      bool: True if the store contains the specified type of attribute
          containers.
    """
    containers = self._attribute_containers.get(container_type, {})
    return bool(containers)

  def Open(self, **kwargs):
    """Opens the store.

    Raises:
      IOError: if the store is already opened.
      OSError: if the store is already opened.
    """
    if self._is_open:
      raise IOError('Store already opened.')

    self._is_open = True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the SQLite-based attribute container store."""

import os
import unittest

from acstore import sqlite_store
from acstore.containers import manager as containers_manager

from tests import test_lib


class _TestSQLiteAttributeContainerStoreV20220716(
    sqlite_store.SQLiteAttributeContainerStore):
  """Test class for testing format compatibility checks."""

  _FORMAT_VERSION = 20220716
  _APPEND_COMPATIBLE_FORMAT_VERSION = 20211121
  _UPGRADE_COMPATIBLE_FORMAT_VERSION = 20211121
  _READ_COMPATIBLE_FORMAT_VERSION = 20211121


class _TestSQLiteAttributeContainerStoreV20221023(
    sqlite_store.SQLiteAttributeContainerStore):
  """Test class for testing format compatibility checks."""

  _FORMAT_VERSION = 20221023
  _APPEND_COMPATIBLE_FORMAT_VERSION = 20221023
  _UPGRADE_COMPATIBLE_FORMAT_VERSION = 20221023
  _READ_COMPATIBLE_FORMAT_VERSION = 20211121


# TODO add tests for PythonAST2SQL.


class SQLiteAttributeContainerStoreTest(test_lib.BaseTestCase):
  """Tests for the SQLite-based storage file object."""

  # pylint: disable=protected-access

  def setUp(self):
    """Sets up the needed objects used throughout the test."""
    containers_manager.AttributeContainersManager.RegisterAttributeContainer(
        test_lib.TestAttributeContainer)

  def tearDown(self):
    """Cleans up the needed objects used throughout the test."""
    containers_manager.AttributeContainersManager.DeregisterAttributeContainer(
        test_lib.TestAttributeContainer)

  def testCacheAttributeContainerByIndex(self):
    """Tests the _CacheAttributeContainerByIndex function."""
    attribute_container = test_lib.TestAttributeContainer()

    with test_lib.TempDirectory():
      test_store = sqlite_store.SQLiteAttributeContainerStore()

      self.assertEqual(len(test_store._attribute_container_cache), 0)

      test_store._CacheAttributeContainerByIndex(attribute_container, 0)
      self.assertEqual(len(test_store._attribute_container_cache), 1)

  def testCheckStorageMetadata(self):
    """Tests the _CheckStorageMetadata function."""
    with test_lib.TempDirectory():
      test_store = sqlite_store.SQLiteAttributeContainerStore()

      metadata_values = {
          'format_version': '{0:d}'.format(test_store._FORMAT_VERSION)}
      test_store._CheckStorageMetadata(metadata_values)

      metadata_values['format_version'] = 'bogus'
      with self.assertRaises(IOError):
        test_store._CheckStorageMetadata(metadata_values)

      metadata_values['format_version'] = '1'
      with self.assertRaises(IOError):
        test_store._CheckStorageMetadata(metadata_values)

      metadata_values['format_version'] = '{0:d}'.format(
          test_store._FORMAT_VERSION)

  def testCreateAttributeContainerTable(self):
    """Tests the _CreateAttributeContainerTable function."""
    attribute_container = test_lib.TestAttributeContainer()

    with test_lib.TempDirectory() as temp_directory:
      test_path = os.path.join(temp_directory, 'acstore.sqlite')
      test_store = sqlite_store.SQLiteAttributeContainerStore()
      test_store.Open(path=test_path, read_only=False)

      try:
        test_store._CreateAttributeContainerTable(
            attribute_container.CONTAINER_TYPE)

        with self.assertRaises(IOError):
          test_store._CreateAttributeContainerTable(
              attribute_container.CONTAINER_TYPE)

      finally:
        test_store.Close()

  # TODO: add tests for _CreatetAttributeContainerFromRow
  # TODO: add tests for _Flush
  # TODO: add tests for _FlushWriteCache

  def testGetAttributeContainersWithFilter(self):
    """Tests the _GetAttributeContainersWithFilter function."""
    attribute_container = test_lib.TestAttributeContainer()
    attribute_container.attribute = '8f0bf95a7959baad9666b21a7feed79d'

    column_names = ['attribute']

    with test_lib.TempDirectory() as temp_directory:
      test_path = os.path.join(temp_directory, 'acstore.sqlite')
      test_store = sqlite_store.SQLiteAttributeContainerStore()
      test_store.Open(path=test_path, read_only=False)

      try:
        containers = list(test_store._GetAttributeContainersWithFilter(
            attribute_container.CONTAINER_TYPE, column_names=column_names))
        self.assertEqual(len(containers), 0)

        test_store.AddAttributeContainer(attribute_container)

        containers = list(test_store._GetAttributeContainersWithFilter(
            attribute_container.CONTAINER_TYPE, column_names=column_names))
        self.assertEqual(len(containers), 1)

        filter_expression = 'attribute == "8f0bf95a7959baad9666b21a7feed79d"'
        containers = list(test_store._GetAttributeContainersWithFilter(
            attribute_container.CONTAINER_TYPE, column_names=column_names,
            filter_expression=filter_expression))
        self.assertEqual(len(containers), 1)

        filter_expression = 'attribute != "8f0bf95a7959baad9666b21a7feed79d"'
        containers = list(test_store._GetAttributeContainersWithFilter(
            attribute_container.CONTAINER_TYPE, column_names=column_names,
            filter_expression=filter_expression))
        self.assertEqual(len(containers), 0)

        containers = list(test_store._GetAttributeContainersWithFilter(
            'bogus', column_names=column_names))
        self.assertEqual(len(containers), 0)

      finally:
        test_store.Close()

  def testGetCachedAttributeContainer(self):
    """Tests the _GetCachedAttributeContainer function."""
    attribute_container = test_lib.TestAttributeContainer()

    with test_lib.TempDirectory():
      test_store = sqlite_store.SQLiteAttributeContainerStore()

      cached_container = test_store._GetCachedAttributeContainer(
          attribute_container.CONTAINER_TYPE, 1)
      self.assertIsNone(cached_container)

      test_store._CacheAttributeContainerByIndex(attribute_container, 1)

      cached_container = test_store._GetCachedAttributeContainer(
          attribute_container.CONTAINER_TYPE, 1)
      self.assertIsNotNone(cached_container)

  def testHasTable(self):
    """Tests the _HasTable function."""
    with test_lib.TempDirectory() as temp_directory:
      test_path = os.path.join(temp_directory, 'acstore.sqlite')
      test_store = sqlite_store.SQLiteAttributeContainerStore()
      test_store.Open(path=test_path, read_only=False)

      try:
        test_store._CreateAttributeContainerTable('test_container')

        result = test_store._HasTable('test_container')
        self.assertTrue(result)

        result = test_store._HasTable('bogus')
        self.assertFalse(result)

      finally:
        test_store.Close()

  def testRaiseIfNotReadable(self):
    """Tests the _RaiseIfNotReadable function."""
    test_store = sqlite_store.SQLiteAttributeContainerStore()

    with self.assertRaises(IOError):
      test_store._RaiseIfNotReadable()

  def testRaiseIfNotWritable(self):
    """Tests the _RaiseIfNotWritable function."""
    test_store = sqlite_store.SQLiteAttributeContainerStore()

    with self.assertRaises(IOError):
      test_store._RaiseIfNotWritable()

  # TODO: add tests for _ReadAndCheckStorageMetadata
  # TODO: add tests for _ReadMetadata
  # TODO: add tests for _UpdateStorageMetadataFormatVersion

  def testWriteExistingAttributeContainer(self):
    """Tests the _WriteExistingAttributeContainer function."""
    attribute_container = test_lib.TestAttributeContainer()

    with test_lib.TempDirectory() as temp_directory:
      test_path = os.path.join(temp_directory, 'acstore.sqlite')
      test_store = sqlite_store.SQLiteAttributeContainerStore()
      test_store.Open(path=test_path, read_only=False)

      try:
        number_of_containers = test_store.GetNumberOfAttributeContainers(
            attribute_container.CONTAINER_TYPE)
        self.assertEqual(number_of_containers, 0)

        test_store._WriteNewAttributeContainer(attribute_container)

        number_of_containers = test_store.GetNumberOfAttributeContainers(
            attribute_container.CONTAINER_TYPE)
        self.assertEqual(number_of_containers, 1)

        test_store._WriteExistingAttributeContainer(attribute_container)

        number_of_containers = test_store.GetNumberOfAttributeContainers(
            attribute_container.CONTAINER_TYPE)
        self.assertEqual(number_of_containers, 1)

      finally:
        test_store.Close()

  # TODO: add tests for _WriteMetadata
  # TODO: add tests for _WriteMetadataValue

  def testWriteNewAttributeContainer(self):
    """Tests the _WriteNewAttributeContainer function."""
    attribute_container = test_lib.TestAttributeContainer()

    with test_lib.TempDirectory() as temp_directory:
      test_path = os.path.join(temp_directory, 'acstore.sqlite')
      test_store = sqlite_store.SQLiteAttributeContainerStore()
      test_store.Open(path=test_path, read_only=False)

      try:
        number_of_containers = test_store.GetNumberOfAttributeContainers(
            attribute_container.CONTAINER_TYPE)
        self.assertEqual(number_of_containers, 0)

        test_store._WriteNewAttributeContainer(attribute_container)

        number_of_containers = test_store.GetNumberOfAttributeContainers(
            attribute_container.CONTAINER_TYPE)
        self.assertEqual(number_of_containers, 1)

      finally:
        test_store.Close()

  def testAddAttributeContainer(self):
    """Tests the AddAttributeContainer function."""
    attribute_container = test_lib.TestAttributeContainer()

    with test_lib.TempDirectory() as temp_directory:
      test_path = os.path.join(temp_directory, 'acstore.sqlite')
      test_store = sqlite_store.SQLiteAttributeContainerStore()
      test_store.Open(path=test_path, read_only=False)

      try:
        number_of_containers = test_store.GetNumberOfAttributeContainers(
            attribute_container.CONTAINER_TYPE)
        self.assertEqual(number_of_containers, 0)

        test_store.AddAttributeContainer(attribute_container)

        number_of_containers = test_store.GetNumberOfAttributeContainers(
            attribute_container.CONTAINER_TYPE)
        self.assertEqual(number_of_containers, 1)

      finally:
        test_store.Close()

      with self.assertRaises(IOError):
        test_store.AddAttributeContainer(attribute_container)

  # TODO: add tests for CheckSupportedFormat

  def testGetAttributeContainerByIdentifier(self):
    """Tests the GetAttributeContainerByIdentifier function."""
    attribute_container = test_lib.TestAttributeContainer()

    with test_lib.TempDirectory() as temp_directory:
      test_path = os.path.join(temp_directory, 'acstore.sqlite')
      test_store = sqlite_store.SQLiteAttributeContainerStore()
      test_store.Open(path=test_path, read_only=False)

      try:
        test_store.AddAttributeContainer(attribute_container)

        identifier = attribute_container.GetIdentifier()

        container = test_store.GetAttributeContainerByIdentifier(
            attribute_container.CONTAINER_TYPE, identifier)
        self.assertIsNotNone(container)

        identifier.sequence_number = 99

        container = test_store.GetAttributeContainerByIdentifier(
            attribute_container.CONTAINER_TYPE, identifier)
        self.assertIsNone(container)

      finally:
        test_store.Close()

  def testGetAttributeContainerByIndex(self):
    """Tests the GetAttributeContainerByIndex function."""
    attribute_container = test_lib.TestAttributeContainer()

    with test_lib.TempDirectory() as temp_directory:
      test_path = os.path.join(temp_directory, 'acstore.sqlite')
      test_store = sqlite_store.SQLiteAttributeContainerStore()
      test_store.Open(path=test_path, read_only=False)

      try:
        container = test_store.GetAttributeContainerByIndex(
            attribute_container.CONTAINER_TYPE, 0)
        self.assertIsNone(container)

        test_store.AddAttributeContainer(attribute_container)

        container = test_store.GetAttributeContainerByIndex(
            attribute_container.CONTAINER_TYPE, 0)
        self.assertIsNotNone(container)

        container = test_store.GetAttributeContainerByIndex('bogus', 0)
        self.assertIsNone(container)

      finally:
        test_store.Close()

  def testGetAttributeContainers(self):
    """Tests the GetAttributeContainers function."""
    attribute_container = test_lib.TestAttributeContainer()
    attribute_container.attribute = '8f0bf95a7959baad9666b21a7feed79d'

    with test_lib.TempDirectory() as temp_directory:
      test_path = os.path.join(temp_directory, 'acstore.sqlite')
      test_store = sqlite_store.SQLiteAttributeContainerStore()
      test_store.Open(path=test_path, read_only=False)

      try:
        containers = list(test_store.GetAttributeContainers(
            attribute_container.CONTAINER_TYPE))
        self.assertEqual(len(containers), 0)

        test_store.AddAttributeContainer(attribute_container)

        containers = list(test_store.GetAttributeContainers(
            attribute_container.CONTAINER_TYPE))
        self.assertEqual(len(containers), 1)

        filter_expression = 'attribute == "8f0bf95a7959baad9666b21a7feed79d"'
        containers = list(test_store.GetAttributeContainers(
            attribute_container.CONTAINER_TYPE,
            filter_expression=filter_expression))
        self.assertEqual(len(containers), 1)

        filter_expression = 'attribute != "8f0bf95a7959baad9666b21a7feed79d"'
        containers = list(test_store.GetAttributeContainers(
            attribute_container.CONTAINER_TYPE,
            filter_expression=filter_expression))
        self.assertEqual(len(containers), 0)

        with self.assertRaises(IOError):
          list(test_store.GetAttributeContainers('bogus'))

      finally:
        test_store.Close()

  def testGetNumberOfAttributeContainers(self):
    """Tests the GetNumberOfAttributeContainers function."""
    attribute_container = test_lib.TestAttributeContainer()

    with test_lib.TempDirectory() as temp_directory:
      test_path = os.path.join(temp_directory, 'acstore.sqlite')
      test_store = sqlite_store.SQLiteAttributeContainerStore()
      test_store.Open(path=test_path, read_only=False)

      try:
        number_of_containers = test_store.GetNumberOfAttributeContainers(
            attribute_container.CONTAINER_TYPE)
        self.assertEqual(number_of_containers, 0)

        test_store.AddAttributeContainer(attribute_container)

        number_of_containers = test_store.GetNumberOfAttributeContainers(
            attribute_container.CONTAINER_TYPE)
        self.assertEqual(number_of_containers, 1)

        # Test for a supported container type that does not have a table
        # present in the storage file.
        query = 'DROP TABLE {0:s}'.format(attribute_container.CONTAINER_TYPE)
        test_store._cursor.execute(query)
        number_of_containers = test_store.GetNumberOfAttributeContainers(
            attribute_container.CONTAINER_TYPE)
        self.assertEqual(number_of_containers, 0)

      finally:
        test_store.Close()

  def testHasAttributeContainers(self):
    """Tests the HasAttributeContainers function."""
    attribute_container = test_lib.TestAttributeContainer()

    with test_lib.TempDirectory() as temp_directory:
      test_path = os.path.join(temp_directory, 'acstore.sqlite')
      test_store = sqlite_store.SQLiteAttributeContainerStore()
      test_store.Open(path=test_path, read_only=False)

      try:
        result = test_store.HasAttributeContainers(
            attribute_container.CONTAINER_TYPE)
        self.assertFalse(result)

        test_store.AddAttributeContainer(attribute_container)

        result = test_store.HasAttributeContainers(
            attribute_container.CONTAINER_TYPE)
        self.assertTrue(result)

        result = test_store.HasAttributeContainers('bogus')
        self.assertFalse(result)

      finally:
        test_store.Close()

  # TODO: add tests for Open and Close

  def testUpdateAttributeContainer(self):
    """Tests the UpdateAttributeContainer function."""
    attribute_container = test_lib.TestAttributeContainer()

    with test_lib.TempDirectory() as temp_directory:
      test_path = os.path.join(temp_directory, 'acstore.sqlite')
      test_store = sqlite_store.SQLiteAttributeContainerStore()
      test_store.Open(path=test_path, read_only=False)

      try:
        number_of_containers = test_store.GetNumberOfAttributeContainers(
            attribute_container.CONTAINER_TYPE)
        self.assertEqual(number_of_containers, 0)

        test_store.AddAttributeContainer(attribute_container)

        number_of_containers = test_store.GetNumberOfAttributeContainers(
            attribute_container.CONTAINER_TYPE)
        self.assertEqual(number_of_containers, 1)

        test_store.UpdateAttributeContainer(attribute_container)

        number_of_containers = test_store.GetNumberOfAttributeContainers(
            attribute_container.CONTAINER_TYPE)
        self.assertEqual(number_of_containers, 1)

      finally:
        test_store.Close()

  def testVersionCompatibility(self):
    """Tests the version compatibility methods."""
    with test_lib.TempDirectory() as temp_directory:
      v1_storage_path = os.path.join(temp_directory, 'v20220716.sqlite')
      v1_test_store = _TestSQLiteAttributeContainerStoreV20220716()
      v1_test_store.Open(path=v1_storage_path, read_only=False)
      v1_test_store.Close()

      v2_test_store_rw = _TestSQLiteAttributeContainerStoreV20221023()

      with self.assertRaises((IOError, OSError)):
        v2_test_store_rw.Open(path=v1_storage_path, read_only=False)

      v2_test_store_ro = _TestSQLiteAttributeContainerStoreV20221023()
      v2_test_store_ro.Open(path=v1_storage_path, read_only=True)
      v2_test_store_ro.Close()


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The profiler classes."""

import codecs
import gzip
import os
import time


class CPUTimeMeasurement(object):
  """The CPU time measurement.

  Attributes:
    start_sample_time (float): start sample time or None if not set.
    total_cpu_time (float): total CPU time or None if not set.
  """

  def __init__(self):
    """Initializes the CPU time measurement."""
    super(CPUTimeMeasurement, self).__init__()
    self._start_cpu_time = None
    self.start_sample_time = None
    self.total_cpu_time = None

  def SampleStart(self):
    """Starts measuring the CPU time."""
    self._start_cpu_time = time.perf_counter()
    self.start_sample_time = time.time()
    self.total_cpu_time = 0

  def SampleStop(self):
    """Stops measuring the CPU time."""
    if self._start_cpu_time is not None:
      self.total_cpu_time += time.perf_counter() - self._start_cpu_time


class StorageProfiler(object):
  """The storage profiler."""

  _FILENAME_PREFIX = 'storage'

  _FILE_HEADER = (
      'Time\tName\tOperation\tDescription\tProcessing time\tData size\t'
      'Compressed data size\n')

  def __init__(self, identifier, path):
    """Initializes a storage profiler.

    Sample files are gzip compressed UTF-8 encoded CSV files.

    Args:
      identifier (str): identifier of the profiling session used to create
          the sample filename.
      path (str): path of the sample file.
    """
    super(StorageProfiler, self).__init__()
    self._identifier = identifier
    self._path = path
    self._profile_measurements = {}
    self._sample_file = None
    self._start_time = None

  def _WritesString(self, content):
    """Writes a string to the sample file.

    Args:
      content (str): content to write to the sample file.
    """
    content_bytes = codecs.encode(content, 'utf-8')
    self._sample_file.write(content_bytes)

  @classmethod
  def IsSupported(cls):
    """Determines if the profiler is supported.

    Returns:
      bool: True if the profiler is supported.
    """
    return True

  def Sample(
      self, profile_name, operation, description, data_size,
      compressed_data_size):
    """Takes a sample of data read or written for profiling.

    Args:
      profile_name (str): name of the profile to sample.
      operation (str): operation, either 'read' or 'write'.
      description (str): description of the data read.
      data_size (int): size of the data read in bytes.
      compressed_data_size (int): size of the compressed data read in bytes.
    """
    measurements = self._profile_measurements.get(profile_name)
    if measurements:
      sample_time = measurements.start_sample_time
      processing_time = measurements.total_cpu_time
    else:
      sample_time = time.time()
      processing_time = 0.0

    sample = '{0:f}\t{1:s}\t{2:s}\t{3:s}\t{4:f}\t{5:d}\t{6:d}\n'.format(
        sample_time, profile_name, operation, description,
        processing_time, data_size, compressed_data_size)
    self._WritesString(sample)

  def Start(self):
    """Starts the profiler."""
    filename = '{0:s}-{1:s}.csv.gz'.format(
        self._FILENAME_PREFIX, self._identifier)
    if self._path:
      filename = os.path.join(self._path, filename)

    self._sample_file = gzip.open(filename, 'wb')
    self._WritesString(self._FILE_HEADER)

    self._start_time = time.time()

  def StartTiming(self, profile_name):
    """Starts timing CPU time.

    Args:
      profile_name (str): name of the profile to sample.
    """
    if profile_name not in self._profile_measurements:
      self._profile_measurements[profile_name] = CPUTimeMeasurement()

    self._profile_measurements[profile_name].SampleStart()

  def Stop(self):
    """Stops the profiler."""
    self._sample_file.close()
    self._sample_file = None

  def StopTiming(self, profile_name):
    """Stops timing CPU time.

    Args:
      profile_name (str): name of the profile to sample.
    """
    measurements = self._profile_measurements.get(profile_name)
    if measurements:
      measurements.SampleStop()



# -*- coding: utf-8 -*-
"""This file contains the attribute container manager class."""


class AttributeContainersManager(object):
  """Class that implements the attribute container manager."""

  _attribute_container_classes = {}

  @classmethod
  def CreateAttributeContainer(cls, container_type):
    """Creates an instance of a specific attribute container type.

    Args:
      container_type (str): container type.

    Returns:
      AttributeContainer: an instance of attribute container.

    Raises:
      ValueError: if the container type is not supported.
    """
    container_class = cls._attribute_container_classes.get(
        container_type, None)
    if not container_class:
      raise ValueError(f'Unsupported container type: {container_type:s}')

    return container_class()

  @classmethod
  def DeregisterAttributeContainer(cls, attribute_container_class):
    """Deregisters an attribute container class.

    The attribute container classes are identified based on their lower case
    container type.

    Args:
      attribute_container_class (type): attribute container class.

    Raises:
      KeyError: if attribute container class is not set for the corresponding
          container type.
    """
    container_type = attribute_container_class.CONTAINER_TYPE.lower()
    if container_type not in cls._attribute_container_classes:
      raise KeyError((
          f'Attribute container class not set for container type: '
          f'{attribute_container_class.CONTAINER_TYPE:s}.'))

    del cls._attribute_container_classes[container_type]

  @classmethod
  def GetContainerTypes(cls):
    """Retrieves the container types of the registered attribute containers.

    Returns:
      list[str]: container types.
    """
    return list(cls._attribute_container_classes.keys())

  @classmethod
  def GetSchema(cls, container_type):
    """Retrieves the schema of a registered attribute container.

    Args:
      container_type (str): attribute container type.

    Returns:
      dict[str, str]: attribute container schema or an empty dictionary if
          no schema available.

    Raises:
      ValueError: if the container type is not supported.
    """
    container_class = cls._attribute_container_classes.get(
        container_type, None)
    if not container_class:
      raise ValueError(f'Unsupported container type: {container_type:s}')

    return getattr(container_class, 'SCHEMA', {})

  @classmethod
  def RegisterAttributeContainer(cls, attribute_container_class):
    """Registers a attribute container class.

    The attribute container classes are identified based on their lower case
    container type.

    Args:
      attribute_container_class (type): attribute container class.

    Raises:
      KeyError: if attribute container class is already set for the
          corresponding container type.
    """
    container_type = attribute_container_class.CONTAINER_TYPE.lower()
    if container_type in cls._attribute_container_classes:
      raise KeyError((
          f'Attribute container class already set for container type: '
          f'{attribute_container_class.CONTAINER_TYPE:s}.'))

    cls._attribute_container_classes[container_type] = attribute_container_class

  @classmethod
  def RegisterAttributeContainers(cls, attribute_container_classes):
    """Registers attribute container classes.

    The attribute container classes are identified based on their lower case
    container type.

    Args:
      attribute_container_classes (list[type]): attribute container classes.

    Raises:
      KeyError: if attribute container class is already set for the
          corresponding container type.
    """
    for attribute_container_class in attribute_container_classes:
      cls.RegisterAttributeContainer(attribute_container_class)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the fake (in-memory only) store."""

import unittest

from acstore import fake_store

from tests import test_lib


class FakeAttributeContainerStoreTest(test_lib.BaseTestCase):
  """Tests for the fake (in-memory only) store."""

  # pylint: disable=protected-access

  def testRaiseIfNotReadable(self):
    """Tests the _RaiseIfNotReadable function."""
    test_store = fake_store.FakeAttributeContainerStore()

    with self.assertRaises(IOError):
      test_store._RaiseIfNotReadable()

  def testRaiseIfNotWritable(self):
    """Tests the _RaiseIfNotWritable function."""
    test_store = fake_store.FakeAttributeContainerStore()

    with self.assertRaises(IOError):
      test_store._RaiseIfNotWritable()

  def testWriteExistingAttributeContainer(self):
    """Tests the _WriteExistingAttributeContainer function."""
    attribute_container = test_lib.TestAttributeContainer()

    test_store = fake_store.FakeAttributeContainerStore()
    test_store.Open()

    number_of_containers = test_store.GetNumberOfAttributeContainers(
        attribute_container.CONTAINER_TYPE)
    self.assertEqual(number_of_containers, 0)

    with self.assertRaises(IOError):
      test_store._WriteExistingAttributeContainer(attribute_container)

    test_store._WriteNewAttributeContainer(attribute_container)

    number_of_containers = test_store.GetNumberOfAttributeContainers(
        attribute_container.CONTAINER_TYPE)
    self.assertEqual(number_of_containers, 1)

    test_store._WriteExistingAttributeContainer(attribute_container)

    number_of_containers = test_store.GetNumberOfAttributeContainers(
        attribute_container.CONTAINER_TYPE)
    self.assertEqual(number_of_containers, 1)

    test_store.Close()

  def testWriteNewAttributeContainer(self):
    """Tests the _WriteNewAttributeContainer function."""
    attribute_container = test_lib.TestAttributeContainer()

    test_store = fake_store.FakeAttributeContainerStore()
    test_store.Open()

    number_of_containers = test_store.GetNumberOfAttributeContainers(
        attribute_container.CONTAINER_TYPE)
    self.assertEqual(number_of_containers, 0)

    test_store._WriteNewAttributeContainer(attribute_container)

    number_of_containers = test_store.GetNumberOfAttributeContainers(
        attribute_container.CONTAINER_TYPE)
    self.assertEqual(number_of_containers, 1)

    test_store.Close()

  def testAddAttributeContainer(self):
    """Tests the AddAttributeContainer function."""
    attribute_container = test_lib.TestAttributeContainer()

    test_store = fake_store.FakeAttributeContainerStore()
    test_store.Open()

    number_of_containers = test_store.GetNumberOfAttributeContainers(
        attribute_container.CONTAINER_TYPE)
    self.assertEqual(number_of_containers, 0)

    test_store.AddAttributeContainer(attribute_container)

    number_of_containers = test_store.GetNumberOfAttributeContainers(
        attribute_container.CONTAINER_TYPE)
    self.assertEqual(number_of_containers, 1)

    test_store.Close()

    with self.assertRaises(IOError):
      test_store.AddAttributeContainer(attribute_container)

  def testGetAttributeContainerByIdentifier(self):
    """Tests the GetAttributeContainerByIdentifier function."""
    attribute_container = test_lib.TestAttributeContainer()

    test_store = fake_store.FakeAttributeContainerStore()
    test_store.Open()

    test_store.AddAttributeContainer(attribute_container)
    identifier = attribute_container.GetIdentifier()

    container = test_store.GetAttributeContainerByIdentifier(
        attribute_container.CONTAINER_TYPE, identifier)
    self.assertIsNotNone(container)

    identifier.sequence_number = 99

    container = test_store.GetAttributeContainerByIdentifier(
        attribute_container.CONTAINER_TYPE, identifier)
    self.assertIsNone(container)

    test_store.Close()

  def testGetAttributeContainerByIndex(self):
    """Tests the GetAttributeContainerByIndex function."""
    attribute_container = test_lib.TestAttributeContainer()

    test_store = fake_store.FakeAttributeContainerStore()
    test_store.Open()

    container = test_store.GetAttributeContainerByIndex(
        attribute_container.CONTAINER_TYPE, 0)
    self.assertIsNone(container)

    test_store.AddAttributeContainer(attribute_container)

    container = test_store.GetAttributeContainerByIndex(
        attribute_container.CONTAINER_TYPE, 0)
    self.assertIsNotNone(container)

    test_store.Close()

  def testGetAttributeContainers(self):
    """Tests the GetAttributeContainers function."""
    attribute_container = test_lib.TestAttributeContainer()
    attribute_container.attribute = '8f0bf95a7959baad9666b21a7feed79d'

    test_store = fake_store.FakeAttributeContainerStore()
    test_store.Open()

    containers = list(test_store.GetAttributeContainers(
        attribute_container.CONTAINER_TYPE))
    self.assertEqual(len(containers), 0)

    test_store.AddAttributeContainer(attribute_container)

    containers = list(test_store.GetAttributeContainers(
        attribute_container.CONTAINER_TYPE))
    self.assertEqual(len(containers), 1)

    filter_expression = 'attribute == "8f0bf95a7959baad9666b21a7feed79d"'
    containers = list(test_store.GetAttributeContainers(
        attribute_container.CONTAINER_TYPE,
        filter_expression=filter_expression))
    self.assertEqual(len(containers), 1)

    filter_expression = 'attribute != "8f0bf95a7959baad9666b21a7feed79d"'
    containers = list(test_store.GetAttributeContainers(
        attribute_container.CONTAINER_TYPE,
        filter_expression=filter_expression))
    self.assertEqual(len(containers), 0)

    test_store.Close()

  def testGetNumberOfAttributeContainers(self):
    """Tests the GetNumberOfAttributeContainers function."""
    attribute_container = test_lib.TestAttributeContainer()

    test_store = fake_store.FakeAttributeContainerStore()
    test_store.Open()

    number_of_containers = test_store.GetNumberOfAttributeContainers(
        attribute_container.CONTAINER_TYPE)
    self.assertEqual(number_of_containers, 0)

    test_store.AddAttributeContainer(attribute_container)

    number_of_containers = test_store.GetNumberOfAttributeContainers(
        attribute_container.CONTAINER_TYPE)
    self.assertEqual(number_of_containers, 1)

    test_store.Close()

  def testHasAttributeContainers(self):
    """Tests the HasAttributeContainers function."""
    attribute_container = test_lib.TestAttributeContainer()

    test_store = fake_store.FakeAttributeContainerStore()
    test_store.Open()

    result = test_store.HasAttributeContainers(
        attribute_container.CONTAINER_TYPE)
    self.assertFalse(result)

    test_store.AddAttributeContainer(attribute_container)

    result = test_store.HasAttributeContainers(
        attribute_container.CONTAINER_TYPE)
    self.assertTrue(result)

    test_store.Close()

  def testOpenClose(self):
    """Tests the Open and Close functions."""
    test_store = fake_store.FakeAttributeContainerStore()
    test_store.Open()
    test_store.Close()

    test_store.Open()
    test_store.Close()

    test_store = fake_store.FakeAttributeContainerStore()
    test_store.Open()
    test_store.Close()

    test_store.Open()

    with self.assertRaises(IOError):
      test_store.Open()

    test_store.Close()

    with self.assertRaises(IOError):
      test_store.Close()

  def testUpdateAttributeContainer(self):
    """Tests the UpdateAttributeContainer function."""
    attribute_container = test_lib.TestAttributeContainer()

    test_store = fake_store.FakeAttributeContainerStore()
    test_store.Open()

    number_of_containers = test_store.GetNumberOfAttributeContainers(
        attribute_container.CONTAINER_TYPE)
    self.assertEqual(number_of_containers, 0)

    with self.assertRaises(IOError):
      test_store.UpdateAttributeContainer(attribute_container)

    test_store.AddAttributeContainer(attribute_container)

    number_of_containers = test_store.GetNumberOfAttributeContainers(
        attribute_container.CONTAINER_TYPE)
    self.assertEqual(number_of_containers, 1)

    test_store.UpdateAttributeContainer(attribute_container)

    number_of_containers = test_store.GetNumberOfAttributeContainers(
        attribute_container.CONTAINER_TYPE)
    self.assertEqual(number_of_containers, 1)

    test_store.Close()


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""Attribute Container Storage (ACStore).

ACStore, or Attribute Container Storage, provides a stand-alone implementation
to read and write plaso storage files.
"""

__version__ = '20230101'



# -*- coding: utf-8 -*-
"""The attribute container interface."""


class AttributeContainerIdentifier(object):
  """The attribute container identifier.

  The identifier is used to uniquely identify attribute containers.
  The value should be unique relative to an attribute container store.

  Attributes:
    name (str): name of the table (attribute container).
    sequence_number (int): sequence number of the attribute container.
  """

  def __init__(self, name=None, sequence_number=None):
    """Initializes an attribute container identifier.

    Args:
      name (Optional[str]): name of the table (attribute container).
      sequence_number (Optional[int]): sequence number of the attribute
          container.
    """
    super(AttributeContainerIdentifier, self).__init__()
    self.name = name
    self.sequence_number = sequence_number

  def CopyFromString(self, identifier_string):
    """Copies the identifier from a string representation.

    Args:
      identifier_string (str): string representation.
    """
    self.name, sequence_number = identifier_string.split('.')
    self.sequence_number = int(sequence_number, 10)

  def CopyToString(self):
    """Copies the identifier to a string representation.

    Returns:
      str: unique identifier or None.
    """
    if self.name is not None and self.sequence_number is not None:
      return '{0:s}.{1:d}'.format(self.name, self.sequence_number)

    return None


class AttributeContainer(object):
  """The attribute container interface.

  This is the the base class for those object that exists primarily as
  a container of attributes with basic accessors and mutators.

  The CONTAINER_TYPE class attribute contains a string that identifies
  the container type, for example the container type "event" identifiers
  an event object.

  Attributes are public class members of an serializable type. Protected and
  private class members are not to be serialized, with the exception of those
  defined in _SERIALIZABLE_PROTECTED_ATTRIBUTES.
  """

  CONTAINER_TYPE = None

  # Names of protected attributes, those with a leading underscore, that
  # should be serialized.
  _SERIALIZABLE_PROTECTED_ATTRIBUTES = []

  def __init__(self):
    """Initializes an attribute container."""
    super(AttributeContainer, self).__init__()
    self._identifier = AttributeContainerIdentifier(
        name=self.CONTAINER_TYPE, sequence_number=id(self))

  def CopyFromDict(self, attributes):
    """Copies the attribute container from a dictionary.

    Args:
      attributes (dict[str, object]): attribute values per name.
    """
    for attribute_name, attribute_value in attributes.items():
      # Not using startswith to improve performance.
      if (attribute_name[0] != '_' or
          attribute_name in self._SERIALIZABLE_PROTECTED_ATTRIBUTES):
        self.__dict__[attribute_name] = attribute_value

  def CopyToDict(self):
    """Copies the attribute container to a dictionary.

    Returns:
      dict[str, object]: attribute values per name.
    """
    return dict(self.GetAttributes())

  def GetAttributeNames(self):
    """Retrieves the names of all attributes.

    Returns:
      list[str]: attribute names.
    """
    attribute_names = list(self._SERIALIZABLE_PROTECTED_ATTRIBUTES)
    for attribute_name in self.__dict__:
      # Not using startswith to improve performance.
      if attribute_name[0] != '_':
        attribute_names.append(attribute_name)

    return attribute_names

  def GetAttributes(self):
    """Retrieves the attribute names and values.

    Attributes that are set to None are ignored.

    Yields:
      tuple[str, object]: attribute name and value.
    """
    for attribute_name, attribute_value in self.__dict__.items():
      # Not using startswith to improve performance.
      if attribute_value is not None and (
          attribute_name[0] != '_' or
          attribute_name in self._SERIALIZABLE_PROTECTED_ATTRIBUTES):
        yield attribute_name, attribute_value

  def GetAttributeValuesHash(self):
    """Retrieves a comparable string of the attribute values.

    Returns:
      int: hash of comparable string of the attribute values.
    """
    return hash(self.GetAttributeValuesString())

  def GetAttributeValuesString(self):
    """Retrieves a comparable string of the attribute values.

    Returns:
      str: comparable string of the attribute values.
    """
    attributes = []
    for attribute_name, attribute_value in sorted(self.__dict__.items()):
      # Not using startswith to improve performance.
      if attribute_value is not None and (
          attribute_name[0] != '_' or
          attribute_name in self._SERIALIZABLE_PROTECTED_ATTRIBUTES):
        if isinstance(attribute_value, dict):
          attribute_value = sorted(attribute_value.items())

        elif isinstance(attribute_value, bytes):
          attribute_value = repr(attribute_value)

        attribute_string = '{0:s}: {1!s}'.format(
            attribute_name, attribute_value)
        attributes.append(attribute_string)

    return ', '.join(attributes)

  def GetIdentifier(self):
    """Retrieves the identifier.

    The identifier is a storage specific value that should not be serialized.

    Returns:
      AttributeContainerIdentifier: an unique identifier for the container.
    """
    return self._identifier

  def MatchesExpression(self, expression):
    """Determines if an attribute container matches the expression.

    Args:
      expression (code|str): expression.

    Returns:
      bool: True if the attribute container matches the expression, False
          otherwise.
    """
    result = not expression
    if expression:
      namespace = {}
      for attribute_name, attribute_value in self.__dict__.items():
        # Not using startswith to improve performance.
        if attribute_value is not None and (
            attribute_name[0] != '_' or
            attribute_name in self._SERIALIZABLE_PROTECTED_ATTRIBUTES):
          if isinstance(attribute_value, AttributeContainerIdentifier):
            attribute_value = attribute_value.CopyToString()

          namespace[attribute_name] = attribute_value

      # Make sure __builtins__ contains an empty dictionary.
      namespace['__builtins__'] = {}

      try:
        result = eval(expression, namespace)  # pylint: disable=eval-used
      except Exception:  # pylint: disable=broad-except
        pass

    return result

  def SetIdentifier(self, identifier):
    """Sets the identifier.

    The identifier is a storage specific value that should not be serialized.

    Args:
      identifier (AttributeContainerIdentifier): identifier.
    """
    self._identifier = identifier



# -*- coding: utf-8 -*-
"""Functions and classes for testing."""

import shutil
import tempfile
import unittest

from acstore.containers import interface as containers_interface


class TestAttributeContainer(containers_interface.AttributeContainer):
  """Attribute container for testing purposes.

  Attributes:
    attribute (str): attribute for testing purposes.
  """

  CONTAINER_TYPE = 'test_container'

  SCHEMA = {'attribute': 'str'}

  def __init__(self):
    """Initializes an attribute container."""
    super(TestAttributeContainer, self).__init__()
    self.attribute = None


class BaseTestCase(unittest.TestCase):
  """The base test case."""

  # Show full diff results.
  maxDiff = None


class TempDirectory(object):
  """Class that implements a temporary directory."""

  def __init__(self):
    """Initializes a temporary directory."""
    super(TempDirectory, self).__init__()
    self.name = ''

  def __enter__(self):
    """Make this work with the 'with' statement."""
    self.name = tempfile.mkdtemp()
    return self.name

  def __exit__(self, exception_type, value, traceback):
    """Make this work with the 'with' statement."""
    shutil.rmtree(self.name, True)



# -*- coding: utf-8 -*-
"""SQLite-based attribute container store."""

import ast
import collections
import itertools
import os
import pathlib
import sqlite3

from acstore.containers import interface as containers_interface
from acstore import interface


def PythonAST2SQL(ast_node):
  """Converts a Python AST to SQL.

  Args:
    ast_node (ast.Node): node of the Python AST.

  Returns:
    str: SQL statement that represents the node.

  Raises:
    TypeError: if the type of node is not supported.
  """
  if isinstance(ast_node, ast.BoolOp):
    if isinstance(ast_node.op, ast.And):
      operand = ' AND '
    elif isinstance(ast_node.op, ast.Or):
      operand = ' OR '
    else:
      raise TypeError(ast_node)

    return operand.join([
        PythonAST2SQL(ast_node_value) for ast_node_value in ast_node.values])

  if isinstance(ast_node, ast.Compare):
    if len(ast_node.ops) != 1:
      raise TypeError(ast_node)

    if isinstance(ast_node.ops[0], ast.Eq):
      operator = ' = '
    elif isinstance(ast_node.ops[0], ast.NotEq):
      operator = ' <> '
    else:
      raise TypeError(ast_node)

    if len(ast_node.comparators) != 1:
      raise TypeError(ast_node)

    sql_left = PythonAST2SQL(ast_node.left)
    sql_right = PythonAST2SQL(ast_node.comparators[0])

    return operator.join([sql_left, sql_right])

  if isinstance(ast_node, ast.Constant):
    if isinstance(ast_node.value, str):
      return f'"{ast_node.value:s}"'

    return str(ast_node.value)

  if isinstance(ast_node, ast.Name):
    return ast_node.id

  if isinstance(ast_node, ast.Num):
    return str(ast_node.n)

  if isinstance(ast_node, ast.Str):
    return f'"{ast_node.s:s}"'

  raise TypeError(ast_node)


class SQLiteAttributeContainerStore(interface.AttributeContainerStore):
  """SQLite-based attribute container store.

  Attributes:
    format_version (int): storage format version.
  """

  _FORMAT_VERSION = 20221023

  # The earliest format version, stored in-file, that this class
  # is able to append (write).
  _APPEND_COMPATIBLE_FORMAT_VERSION = 20221023

  # The earliest format version, stored in-file, that this class
  # is able to upgrade (write new format features).
  _UPGRADE_COMPATIBLE_FORMAT_VERSION = 20221023

  # The earliest format version, stored in-file, that this class
  # is able to read.
  _READ_COMPATIBLE_FORMAT_VERSION = 20221023

  _CONTAINER_SCHEMA_TO_SQLITE_TYPE_MAPPINGS = {
      'AttributeContainerIdentifier': 'TEXT',
      'bool': 'INTEGER',
      'int': 'INTEGER',
      'str': 'TEXT',
      'timestamp': 'BIGINT'}

  _CREATE_METADATA_TABLE_QUERY = (
      'CREATE TABLE metadata (key TEXT, value TEXT);')

  _HAS_TABLE_QUERY = (
      'SELECT name FROM sqlite_master '
      'WHERE type = "table" AND name = "{0:s}"')

  _INSERT_METADATA_VALUE_QUERY = (
      'INSERT INTO metadata (key, value) VALUES (?, ?)')

  # The maximum number of cached attribute containers
  _MAXIMUM_CACHED_CONTAINERS = 32 * 1024

  _MAXIMUM_WRITE_CACHE_SIZE = 50

  def __init__(self):
    """Initializes a SQLite attribute container store."""
    super(SQLiteAttributeContainerStore, self).__init__()
    self._attribute_container_cache = collections.OrderedDict()
    self._connection = None
    self._cursor = None
    self._is_open = False
    self._read_only = True
    self._write_cache = {}

    self.format_version = self._FORMAT_VERSION

  def _CacheAttributeContainerByIndex(self, attribute_container, index):
    """Caches a specific attribute container.

    Args:
      attribute_container (AttributeContainer): attribute container.
      index (int): attribute container index.
    """
    if len(self._attribute_container_cache) >= self._MAXIMUM_CACHED_CONTAINERS:
      self._attribute_container_cache.popitem(last=True)

    lookup_key = f'{attribute_container.CONTAINER_TYPE:s}.{index:d}'
    self._attribute_container_cache[lookup_key] = attribute_container
    self._attribute_container_cache.move_to_end(lookup_key, last=False)

  def _CacheAttributeContainerForWrite(
      self, container_type, column_names, values):
    """Caches an attribute container for writing.

    Args:
      container_type (str): attribute container type.
      column_names (list[str]): names of the columns.
      values (list[str]): values for each of the colums.
    """
    write_cache = self._write_cache.get(container_type, [column_names])
    write_cache.append(values)

    if len(write_cache) >= self._MAXIMUM_WRITE_CACHE_SIZE:
      self._FlushWriteCache(container_type, write_cache)
      write_cache = [column_names]

    self._write_cache[container_type] = write_cache

  def _CheckStorageMetadata(self, metadata_values, check_readable_only=False):
    """Checks the storage metadata.

    Args:
      metadata_values (dict[str, str]): metadata values per key.
      check_readable_only (Optional[bool]): whether the store should only be
          checked to see if it can be read. If False, the store will be checked
          to see if it can be read and written to.

    Raises:
      IOError: if the format version is not supported.
      OSError: if the format version is not supported.
    """
    format_version = metadata_values.get('format_version', None)

    if not format_version:
      raise IOError('Missing format version.')

    try:
      format_version = int(format_version, 10)
    except (TypeError, ValueError):
      raise IOError(f'Invalid format version: {format_version!s}.')

    if (not check_readable_only and
        format_version < self._APPEND_COMPATIBLE_FORMAT_VERSION):
      raise IOError((
          f'Format version: {format_version:d} is too old and can no longer '
          f'be written.'))

    if format_version < self._READ_COMPATIBLE_FORMAT_VERSION:
      raise IOError((
          f'Format version: {format_version:d} is too old and can no longer '
          f'be read.'))

    if format_version > self._FORMAT_VERSION:
      raise IOError((
          f'Format version: {format_version:d} is too new and not yet '
          f'supported.'))

    metadata_values['format_version'] = format_version

  def _CommitWriteCache(self, container_type):
    """Commits the write cache for a specific type of attribute container.

    Args:
      container_type (str): attribute container type.
    """
    write_cache = self._write_cache.get(container_type, [])
    if len(write_cache) > 1:
      self._FlushWriteCache(container_type, write_cache)
      del self._write_cache[container_type]

  def _CreateAttributeContainerTable(self, container_type):
    """Creates a table for a specific attribute container type.

    Args:
      container_type (str): attribute container type.

    Raises:
      IOError: when there is an error querying the attribute container store
          or if an unsupported attribute container is provided.
      OSError: when there is an error querying the attribute container store
          or if an unsupported attribute container is provided.
    """
    schema = self._GetAttributeContainerSchema(container_type)
    if not schema:
      raise IOError(f'Unsupported attribute container type: {container_type:s}')

    column_definitions = ['_identifier INTEGER PRIMARY KEY AUTOINCREMENT']

    schema_to_sqlite_type_mappings = (
          self._CONTAINER_SCHEMA_TO_SQLITE_TYPE_MAPPINGS)
    for name, data_type in sorted(schema.items()):
      data_type = schema_to_sqlite_type_mappings.get(data_type, 'TEXT')
      column_definitions.append(f'{name:s} {data_type:s}')

    column_definitions = ', '.join(column_definitions)
    query = f'CREATE TABLE {container_type:s} ({column_definitions:s});'

    try:
      self._cursor.execute(query)
    except (sqlite3.InterfaceError, sqlite3.OperationalError) as exception:
      raise IOError((
          f'Unable to query attribute container store with error: '
          f'{exception!s}'))

  def _CreatetAttributeContainerFromRow(
      self, container_type, column_names, row, first_column_index):
    """Creates an attribute container of a row in the database.

    Args:
      container_type (str): attribute container type.
      column_names (list[str]): names of the columns selected.
      row (sqlite.Row): row as a result from a SELECT query.
      first_column_index (int): index of the first column in row.

    Returns:
      AttributeContainer: attribute container.

    Raises:
      IOError: when there is an error querying the attribute container store
          or if an unsupported attribute container is provided.
      OSError: when there is an error querying the attribute container store
          or if an unsupported attribute container is provided.
    """
    schema = self._GetAttributeContainerSchema(container_type)
    if not schema:
      raise IOError(f'Unsupported attribute container type: {container_type:s}')

    container = self._containers_manager.CreateAttributeContainer(
        container_type)

    for column_index, name in enumerate(column_names):
      attribute_value = row[first_column_index + column_index]
      if attribute_value is None:
        continue

      data_type = schema[name]
      if data_type == 'AttributeContainerIdentifier':
        identifier = containers_interface.AttributeContainerIdentifier()
        identifier.CopyFromString(attribute_value)
        attribute_value = identifier

      elif data_type == 'bool':
        attribute_value = bool(attribute_value)

      elif data_type not in self._CONTAINER_SCHEMA_TO_SQLITE_TYPE_MAPPINGS:
        raise IOError((
            f'Unsupported attribute container type: {container_type:s} '
            f'attribute: {name:s} data type: {data_type:s}'))

      setattr(container, name, attribute_value)

    return container

  def _Flush(self):
    """Ensures cached data is written to file.

    Raises:
      IOError: when there is an error querying the attribute container store.
      OSError: when there is an error querying the attribute container store.
    """
    for container_type, write_cache in self._write_cache.items():
      if len(write_cache) > 1:
        self._FlushWriteCache(container_type, write_cache)

    self._write_cache = {}

    # We need to run commit or not all data is stored in the database.
    self._connection.commit()

  def _FlushWriteCache(self, container_type, write_cache):
    """Flushes attribute container values cached for writing.

    Args:
      container_type (str): attribute container type.
      write_cache (list[tuple[str]]): cached attribute container values.

    Raises:
      IOError: when there is an error querying the attribute container store.
      OSError: when there is an error querying the attribute container store.
    """
    column_names = write_cache.pop(0)

    value_statement = ','.join(['?'] * len(column_names))
    value_statement = f'({value_statement:s})'
    values_statement = ', '.join([value_statement] * len(write_cache))

    column_names_string = ', '.join(column_names)

    query = (f'INSERT INTO {container_type:s} ({column_names_string:s}) '
             f'VALUES {values_statement:s}')

    if self._storage_profiler:
      self._storage_profiler.StartTiming('write_new')

    try:
      values = list(itertools.chain(*write_cache))
      self._cursor.execute(query, values)

    except (sqlite3.InterfaceError, sqlite3.OperationalError) as exception:
      raise IOError((
          f'Unable to query attribute container store with error: '
          f'{exception!s}'))

    finally:
      if self._storage_profiler:
        self._storage_profiler.StopTiming('write_new')

  def _GetAttributeContainersWithFilter(
      self, container_type, column_names=None, filter_expression=None,
      order_by=None):
    """Retrieves a specific type of stored attribute containers.

    Args:
      container_type (str): attribute container type.
      column_names (Optional[list[str]]): names of the columns to retrieve.
      filter_expression (Optional[str]): SQL expression to filter results by.
      order_by (Optional[str]): name of a column to order the results by.

    Yields:
      AttributeContainer: attribute container.

    Raises:
      IOError: when there is an error querying the attribute container store.
      OSError: when there is an error querying the attribute container store.
    """
    self._CommitWriteCache(container_type)

    if self._attribute_container_sequence_numbers[container_type]:
      column_names_string = ', '.join(column_names)

      query = (f'SELECT _identifier, {column_names_string:s} '
               f'FROM {container_type:s}')
      if filter_expression:
        query = ' WHERE '.join([query, filter_expression])
      if order_by:
        query = ' ORDER BY '.join([query, order_by])

      # Use a local cursor to prevent another query interrupting the generator.
      cursor = self._connection.cursor()

      try:
        cursor.execute(query)
      except (sqlite3.InterfaceError, sqlite3.OperationalError) as exception:
        raise IOError((
            f'Unable to query attribute container store for container: '
            f'{container_type:s} with error: {exception!s}'))

      if self._storage_profiler:
        self._storage_profiler.StartTiming('get_containers')

      try:
        row = cursor.fetchone()

      finally:
        if self._storage_profiler:
          self._storage_profiler.StopTiming('get_containers')

      while row:
        container = self._CreatetAttributeContainerFromRow(
            container_type, column_names, row, 1)

        identifier = containers_interface.AttributeContainerIdentifier(
            name=container_type, sequence_number=row[0])
        container.SetIdentifier(identifier)

        yield container

        if self._storage_profiler:
          self._storage_profiler.StartTiming('get_containers')

        try:
          row = cursor.fetchone()

        finally:
          if self._storage_profiler:
            self._storage_profiler.StopTiming('get_containers')

  def _GetCachedAttributeContainer(self, container_type, index):
    """Retrieves a specific cached attribute container.

    Args:
      container_type (str): attribute container type.
      index (int): attribute container index.

    Returns:
      AttributeContainer: attribute container or None if not available.

    Raises:
      IOError: when there is an error querying the attribute container store.
      OSError: when there is an error querying the attribute container store.
    """
    lookup_key = f'{container_type:s}.{index:d}'
    attribute_container = self._attribute_container_cache.get(lookup_key, None)
    if attribute_container:
      self._attribute_container_cache.move_to_end(lookup_key, last=False)
    return attribute_container

  def _HasTable(self, table_name):
    """Determines if a specific table exists.

    Args:
      table_name (str): name of the table.

    Returns:
      bool: True if the table exists, false otherwise.

    Raises:
      IOError: when there is an error querying the attribute container store.
      OSError: when there is an error querying the attribute container store.
    """
    query = self._HAS_TABLE_QUERY.format(table_name)

    try:
      self._cursor.execute(query)
    except (sqlite3.InterfaceError, sqlite3.OperationalError) as exception:
      raise IOError((
          f'Unable to query attribute container store with error: '
          f'{exception!s}'))

    return bool(self._cursor.fetchone())

  def _RaiseIfNotReadable(self):
    """Raises if the attribute container store is not readable.

     Raises:
      IOError: when the attribute container store is closed.
      OSError: when the attribute container store is closed.
    """
    if not self._is_open:
      raise IOError('Unable to read from closed attribute container store.')

  def _RaiseIfNotWritable(self):
    """Raises if the attribute container store is not writable.

    Raises:
      IOError: when the attribute container store is closed or read-only.
      OSError: when the attribute container store is closed or read-only.
    """
    if not self._is_open:
      raise IOError('Unable to write to closed attribute container store.')

    if self._read_only:
      raise IOError('Unable to write to read-only attribute container store.')

  def _ReadAndCheckStorageMetadata(self, check_readable_only=False):
    """Reads storage metadata and checks that the values are valid.

    Args:
      check_readable_only (Optional[bool]): whether the store should only be
          checked to see if it can be read. If False, the store will be checked
          to see if it can be read and written to.

    Raises:
      IOError: when there is an error querying the attribute container store.
      OSError: when there is an error querying the attribute container store.
    """
    metadata_values = self._ReadMetadata()

    self._CheckStorageMetadata(
        metadata_values, check_readable_only=check_readable_only)

    self.format_version = metadata_values['format_version']

  def _ReadMetadata(self):
    """Reads metadata.

    Returns:
      dict[str, str]: metadata values.

    Raises:
      IOError: when there is an error querying the attribute container store.
      OSError: when there is an error querying the attribute container store.
    """
    query = 'SELECT key, value FROM metadata'

    try:
      self._cursor.execute(query)
    except (sqlite3.InterfaceError, sqlite3.OperationalError) as exception:
      raise IOError((
          f'Unable to query attribute container store with error: '
          f'{exception!s}'))

    return {row[0]: row[1] for row in self._cursor.fetchall()}

  def _UpdateStorageMetadataFormatVersion(self):
    """Updates the storage metadata format version.

    Raises:
      IOError: when there is an error querying the attribute container store.
      OSError: when there is an error querying the attribute container store.
    """
    if self.format_version >= self._UPGRADE_COMPATIBLE_FORMAT_VERSION:
      query = (f'UPDATE metadata SET value = {self._FORMAT_VERSION:d} '
               f'WHERE key = "format_version"')

      try:
        self._cursor.execute(query)
      except (sqlite3.InterfaceError, sqlite3.OperationalError) as exception:
        raise IOError((
            f'Unable to query attribute container store with error: '
            f'{exception!s}'))

  def _WriteExistingAttributeContainer(self, container):
    """Writes an existing attribute container to the store.

    The table for the container type must exist.

    Args:
      container (AttributeContainer): attribute container.

    Raises:
      IOError: when there is an error querying the attribute container store
          or if an unsupported attribute container is provided.
      OSError: when there is an error querying the attribute container store
          or if an unsupported attribute container is provided.
    """
    self._CommitWriteCache(container.CONTAINER_TYPE)

    identifier = container.GetIdentifier()

    schema = self._GetAttributeContainerSchema(container.CONTAINER_TYPE)
    if not schema:
      raise IOError(
          f'Unsupported attribute container type: {container.CONTAINER_TYPE:s}')

    column_names = []
    values = []
    for name, data_type in sorted(schema.items()):
      attribute_value = getattr(container, name, None)
      if attribute_value is not None:
        if data_type == 'AttributeContainerIdentifier' and isinstance(
            attribute_value, containers_interface.AttributeContainerIdentifier):
          attribute_value = attribute_value.CopyToString()

        elif data_type == 'bool':
          attribute_value = int(attribute_value)

        elif data_type not in self._CONTAINER_SCHEMA_TO_SQLITE_TYPE_MAPPINGS:
          raise IOError((
              f'Unsupported attribute container type: '
              f'{container.CONTAINER_TYPE:s} attribute: {name:s} data type: '
              f'{data_type:s}'))

      column_names.append(f'{name:s} = ?')
      values.append(attribute_value)

    column_names_string = ', '.join(column_names)

    query = (f'UPDATE {container.CONTAINER_TYPE:s} SET {column_names_string:s} '
             f'WHERE _identifier = {identifier.sequence_number:d}')

    if self._storage_profiler:
      self._storage_profiler.StartTiming('write_existing')

    try:
      self._cursor.execute(query, values)

    except (sqlite3.InterfaceError, sqlite3.OperationalError) as exception:
      raise IOError((
          f'Unable to query attribute container store with error: '
          f'{exception!s}'))

    finally:
      if self._storage_profiler:
        self._storage_profiler.StopTiming('write_existing')

  def _WriteMetadata(self):
    """Writes metadata.

    Raises:
      IOError: when there is an error querying the attribute container store.
      OSError: when there is an error querying the attribute container store.
    """
    try:
      self._cursor.execute(self._CREATE_METADATA_TABLE_QUERY)
    except (sqlite3.InterfaceError, sqlite3.OperationalError) as exception:
      raise IOError((
          f'Unable to query attribute container store with error: '
          f'{exception!s}'))

    self._WriteMetadataValue('format_version', f'{self._FORMAT_VERSION:d}')

  def _WriteMetadataValue(self, key, value):
    """Writes a metadata value.

    Args:
      key (str): key of the storage metadata.
      value (str): value of the storage metadata.

    Raises:
      IOError: when there is an error querying the attribute container store.
      OSError: when there is an error querying the attribute container store.
    """
    try:
      self._cursor.execute(self._INSERT_METADATA_VALUE_QUERY, (key, value))
    except (sqlite3.InterfaceError, sqlite3.OperationalError) as exception:
      raise IOError((
          f'Unable to query attribute container store with error: '
          f'{exception!s}'))

  def _WriteNewAttributeContainer(self, container):
    """Writes a new attribute container to the store.

    The table for the container type is created if needed.

    Args:
      container (AttributeContainer): attribute container.

    Raises:
      IOError: when there is an error querying the attribute container store
          or if an unsupported attribute container is provided.
      OSError: when there is an error querying the attribute container store
          or if an unsupported attribute container is provided.
    """
    next_sequence_number = self._GetAttributeContainerNextSequenceNumber(
        container.CONTAINER_TYPE)

    if (next_sequence_number == 1 and
        not self._HasTable(container.CONTAINER_TYPE)):
      self._CreateAttributeContainerTable(container.CONTAINER_TYPE)

    identifier = containers_interface.AttributeContainerIdentifier(
        name=container.CONTAINER_TYPE, sequence_number=next_sequence_number)
    container.SetIdentifier(identifier)

    schema = self._GetAttributeContainerSchema(container.CONTAINER_TYPE)
    if not schema:
      raise IOError(
          f'Unsupported attribute container type: {container.CONTAINER_TYPE:s}')

    column_names = []
    values = []
    for name, data_type in sorted(schema.items()):
      attribute_value = getattr(container, name, None)
      if attribute_value is not None:
        if data_type == 'AttributeContainerIdentifier' and isinstance(
            attribute_value,
            containers_interface.AttributeContainerIdentifier):
          attribute_value = attribute_value.CopyToString()

        elif data_type == 'bool':
          attribute_value = int(attribute_value)

        elif data_type not in self._CONTAINER_SCHEMA_TO_SQLITE_TYPE_MAPPINGS:
          raise IOError((
              f'Unsupported attribute container type: '
              f'{container.CONTAINER_TYPE:s} attribute: {name:s} data type: '
              f'{data_type:s}'))

      column_names.append(name)
      values.append(attribute_value)

    self._CacheAttributeContainerForWrite(
        container.CONTAINER_TYPE, column_names, values)

    self._CacheAttributeContainerByIndex(container, next_sequence_number - 1)

  @classmethod
  def CheckSupportedFormat(cls, path):
    """Checks if the attribute container store format is supported.

    Args:
      path (str): path to the attribute container store.

    Returns:
      bool: True if the format is supported.
    """
    # Check if the path is an existing file, to prevent sqlite3 creating
    # an emtpy database file.
    if not os.path.isfile(path):
      return False

    try:
      connection = sqlite3.connect(
          path, detect_types=sqlite3.PARSE_DECLTYPES|sqlite3.PARSE_COLNAMES)

      cursor = connection.cursor()

      query = 'SELECT * FROM metadata'
      cursor.execute(query)

      metadata_values = {row[0]: row[1] for row in cursor.fetchall()}

      format_version = metadata_values.get('format_version', None)
      if format_version:
        try:
          format_version = int(format_version, 10)
          result = True
        except (TypeError, ValueError):
          pass

      connection.close()

    except (IOError, TypeError, ValueError, sqlite3.DatabaseError):
      result = False

    return result

  def Close(self):
    """Closes the file.

    Raises:
      IOError: if the attribute container store is already closed.
      OSError: if the attribute container store is already closed.
    """
    if not self._is_open:
      raise IOError('Storage file already closed.')

    if self._connection:
      self._Flush()

      self._connection.close()

      self._connection = None
      self._cursor = None

    self._is_open = False

  def GetAttributeContainerByIdentifier(self, container_type, identifier):
    """Retrieves a specific type of container with a specific identifier.

    Args:
      container_type (str): container type.
      identifier (AttributeContainerIdentifier): attribute container identifier.

    Returns:
      AttributeContainer: attribute container or None if not available.

    Raises:
      IOError: when the store is closed or if an unsupported attribute
          container is provided.
      OSError: when the store is closed or if an unsupported attribute
          container is provided.
    """
    return self.GetAttributeContainerByIndex(
        container_type, identifier.sequence_number - 1)

  def GetAttributeContainerByIndex(self, container_type, index):
    """Retrieves a specific attribute container.

    Args:
      container_type (str): attribute container type.
      index (int): attribute container index.

    Returns:
      AttributeContainer: attribute container or None if not available.

    Raises:
      IOError: when there is an error querying the attribute container store
          or if an unsupported attribute container is provided.
      OSError: when there is an error querying the attribute container store
          or if an unsupported attribute container is provided.
    """
    container = self._GetCachedAttributeContainer(container_type, index)
    if container:
      return container

    self._CommitWriteCache(container_type)

    if not self._attribute_container_sequence_numbers[container_type]:
      return None

    schema = self._GetAttributeContainerSchema(container_type)
    if not schema:
      raise IOError(f'Unsupported attribute container type: {container_type:s}')

    column_names = sorted(schema.keys())

    column_names_string = ', '.join(column_names)
    row_number = index + 1

    query = (f'SELECT {column_names_string:s} FROM {container_type:s} WHERE '
             f'rowid = {row_number:d}')

    try:
      self._cursor.execute(query)
    except (sqlite3.InterfaceError, sqlite3.OperationalError) as exception:
      raise IOError((
          f'Unable to query attribute container store with error: '
          f'{exception!s}'))

    if self._storage_profiler:
      self._storage_profiler.StartTiming('get_container_by_index')

    try:
      row = self._cursor.fetchone()

    finally:
      if self._storage_profiler:
        self._storage_profiler.StopTiming('get_container_by_index')

    if not row:
      return None

    container = self._CreatetAttributeContainerFromRow(
        container_type, column_names, row, 0)

    identifier = containers_interface.AttributeContainerIdentifier(
        name=container_type, sequence_number=row_number)
    container.SetIdentifier(identifier)

    self._CacheAttributeContainerByIndex(container, index)
    return container

  def GetAttributeContainers(self, container_type, filter_expression=None):
    """Retrieves a specific type of stored attribute containers.

    Args:
      container_type (str): attribute container type.
      filter_expression (Optional[str]): expression to filter the resulting
          attribute containers by.

    Returns:
      generator(AttributeContainer): attribute container generator.

    Raises:
      IOError: when there is an error querying the attribute container store
          or if an unsupported attribute container is provided.
      OSError: when there is an error querying the attribute container store
          or if an unsupported attribute container is provided.
    """
    schema = self._GetAttributeContainerSchema(container_type)
    if not schema:
      raise IOError(f'Unsupported attribute container type: {container_type:s}')

    column_names = sorted(schema.keys())

    sql_filter_expression = None
    if filter_expression:
      expression_ast = ast.parse(filter_expression, mode='eval')
      sql_filter_expression = PythonAST2SQL(expression_ast.body)

    return self._GetAttributeContainersWithFilter(
        container_type, column_names=column_names,
        filter_expression=sql_filter_expression)

  def GetNumberOfAttributeContainers(self, container_type):
    """Retrieves the number of a specific type of attribute containers.

    Args:
      container_type (str): attribute container type.

    Returns:
      int: the number of containers of a specified type.

    Raises:
      IOError: when there is an error querying the attribute container store.
      OSError: when there is an error querying the attribute container store.
    """
    self._CommitWriteCache(container_type)

    if not self._HasTable(container_type):
      return 0

    # Note that this is SQLite specific, and will give inaccurate results if
    # there are DELETE commands run on the table. acstore does not run any
    # DELETE commands.
    query = f'SELECT MAX(_ROWID_) FROM {container_type:s} LIMIT 1'

    try:
      self._cursor.execute(query)
    except (sqlite3.InterfaceError, sqlite3.OperationalError) as exception:
      raise IOError((
          f'Unable to query attribute container store with error: '
          f'{exception!s}'))

    row = self._cursor.fetchone()
    if not row:
      return 0

    return row[0] or 0

  def HasAttributeContainers(self, container_type):
    """Determines if store contains a specific type of attribute containers.

    Args:
      container_type (str): attribute container type.

    Returns:
      bool: True if the store contains the specified type of attribute
          containers.

    Raises:
      IOError: when there is an error querying the attribute container store.
      OSError: when there is an error querying the attribute container store.
    """
    count = self.GetNumberOfAttributeContainers(container_type)
    return count > 0

  def Open(self, path=None, read_only=True, **unused_kwargs):  # pylint: disable=arguments-differ
    """Opens the store.

    Args:
      path (Optional[str]): path to the attribute container store.
      read_only (Optional[bool]): True if the file should be opened in
          read-only mode.

    Raises:
      IOError: if the attribute container store is already opened or if
          the database cannot be connected.
      OSError: if the attribute container store is already opened or if
          the database cannot be connected.
      ValueError: if path is missing.
    """
    if self._is_open:
      raise IOError('Storage file already opened.')

    if not path:
      raise ValueError('Missing path.')

    path = os.path.abspath(path)

    try:
      path_uri = pathlib.Path(path).as_uri()
      if read_only:
        path_uri = f'{path_uri:s}?mode=ro'

    except ValueError:
      path_uri = None

    detect_types = sqlite3.PARSE_DECLTYPES|sqlite3.PARSE_COLNAMES

    if path_uri:
      connection = sqlite3.connect(
          path_uri, detect_types=detect_types, isolation_level='DEFERRED',
          uri=True)
    else:
      connection = sqlite3.connect(
          path, detect_types=detect_types, isolation_level='DEFERRED')

    try:
      # Use in-memory journaling mode to reduce IO.
      connection.execute('PRAGMA journal_mode=MEMORY')

      # Turn off insert transaction integrity since we want to do bulk insert.
      connection.execute('PRAGMA synchronous=OFF')

    except (sqlite3.InterfaceError, sqlite3.OperationalError) as exception:
      raise IOError((
          f'Unable to query attribute container store with error: '
          f'{exception!s}'))

    cursor = connection.cursor()
    if not cursor:
      return

    self._connection = connection
    self._cursor = cursor
    self._is_open = True
    self._read_only = read_only

    if read_only:
      self._ReadAndCheckStorageMetadata(check_readable_only=True)
    else:
      if not self._HasTable('metadata'):
        self._WriteMetadata()
      else:
        self._ReadAndCheckStorageMetadata()

        # Update the storage metadata format version in case we are adding
        # new format features that are not backwards compatible.
        self._UpdateStorageMetadataFormatVersion()

      self._connection.commit()

    # Initialize next_sequence_number based on the file contents so that
    # AttributeContainerIdentifier points to the correct attribute container.
    for container_type in self._containers_manager.GetContainerTypes():
      next_sequence_number = self.GetNumberOfAttributeContainers(container_type)
      self._SetAttributeContainerNextSequenceNumber(
          container_type, next_sequence_number)



# -*- coding: utf-8 -*-
"""The attribute container store interface."""

import abc
import collections

from acstore.containers import manager as containers_manager


class AttributeContainerStore(object):
  """Interface of an attribute container store.

  Attributes:
    format_version (int): storage format version.
  """

  def __init__(self):
    """Initializes an attribute container store."""
    super(AttributeContainerStore, self).__init__()
    self._attribute_container_sequence_numbers = collections.Counter()
    self._containers_manager = containers_manager.AttributeContainersManager
    self._storage_profiler = None

    self.format_version = None

  def _GetAttributeContainerNextSequenceNumber(self, container_type):
    """Retrieves the next sequence number of an attribute container.

    Args:
      container_type (str): attribute container type.

    Returns:
      int: next sequence number.
    """
    self._attribute_container_sequence_numbers[container_type] += 1
    return self._attribute_container_sequence_numbers[container_type]

  def _GetAttributeContainerSchema(self, container_type):
    """Retrieves the schema of an attribute container.

    Args:
      container_type (str): attribute container type.

    Returns:
      dict[str, str]: attribute container schema or an empty dictionary if
          no schema available.
    """
    try:
      schema = self._containers_manager.GetSchema(container_type)
    except ValueError:
      schema = {}

    return schema

  @abc.abstractmethod
  def _RaiseIfNotReadable(self):
    """Raises if the store is not readable.

     Raises:
       OSError: if the store cannot be read from.
       IOError: if the store cannot be read from.
    """

  @abc.abstractmethod
  def _RaiseIfNotWritable(self):
    """Raises if the store is not writable.

     Raises:
       OSError: if the store cannot be written to.
       IOError: if the store cannot be written to.
    """

  def _SetAttributeContainerNextSequenceNumber(
      self, container_type, next_sequence_number):
    """Sets the next sequence number of an attribute container.

    Args:
      container_type (str): attribute container type.
      next_sequence_number (int): next sequence number.
    """
    self._attribute_container_sequence_numbers[
        container_type] = next_sequence_number

  @abc.abstractmethod
  def _WriteNewAttributeContainer(self, container):
    """Writes a new attribute container to the store.

    Args:
      container (AttributeContainer): attribute container.
    """

  @abc.abstractmethod
  def _WriteExistingAttributeContainer(self, container):
    """Writes an existing attribute container to the store.

    Args:
      container (AttributeContainer): attribute container.
    """

  def AddAttributeContainer(self, container):
    """Adds a new attribute container.

    Args:
      container (AttributeContainer): attribute container.

    Raises:
      OSError: if the store cannot be written to.
      IOError: if the store cannot be written to.
    """
    self._RaiseIfNotWritable()
    self._WriteNewAttributeContainer(container)

  @abc.abstractmethod
  def Close(self):
    """Closes the store."""

  @abc.abstractmethod
  def GetAttributeContainerByIdentifier(self, container_type, identifier):
    """Retrieves a specific type of container with a specific identifier.

    Args:
      container_type (str): container type.
      identifier (AttributeContainerIdentifier): attribute container identifier.

    Returns:
      AttributeContainer: attribute container or None if not available.

    Raises:
      IOError: when the store is closed or if an unsupported identifier is
          provided.
      OSError: when the store is closed or if an unsupported identifier is
          provided.
    """

  @abc.abstractmethod
  def GetAttributeContainerByIndex(self, container_type, index):
    """Retrieves a specific attribute container.

    Args:
      container_type (str): attribute container type.
      index (int): attribute container index.

    Returns:
      AttributeContainer: attribute container or None if not available.

    Raises:
      IOError: when the store is closed.
      OSError: when the store is closed.
    """

  @abc.abstractmethod
  def GetAttributeContainers(self, container_type, filter_expression=None):
    """Retrieves a specific type of attribute containers.

    Args:
      container_type (str): attribute container type.
      filter_expression (Optional[str]): expression to filter the resulting
          attribute containers by.

    Returns:
      generator(AttributeContainer): attribute container generator.

    Raises:
      IOError: when the store is closed.
      OSError: when the store is closed.
    """

  @abc.abstractmethod
  def GetNumberOfAttributeContainers(self, container_type):
    """Retrieves the number of a specific type of attribute containers.

    Args:
      container_type (str): attribute container type.

    Returns:
      int: the number of containers of a specified type.
    """

  @abc.abstractmethod
  def HasAttributeContainers(self, container_type):
    """Determines if a store contains a specific type of attribute container.

    Args:
      container_type (str): attribute container type.

    Returns:
      bool: True if the store contains the specified type of attribute
          containers.
    """

  @abc.abstractmethod
  def Open(self, **kwargs):
    """Opens the store."""

  def SetStorageProfiler(self, storage_profiler):
    """Sets the storage profiler.

    Args:
      storage_profiler (StorageProfiler): storage profiler.
    """
    self._storage_profiler = storage_profiler

  def UpdateAttributeContainer(self, container):
    """Updates an existing attribute container.

    Args:
      container (AttributeContainer): attribute container.

    Raises:
      OSError: if the store cannot be written to.
      IOError: if the store cannot be written to.
    """
    self._RaiseIfNotWritable()
    self._WriteExistingAttributeContainer(container)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the attribute container interface."""

import unittest

from acstore.containers import interface

from tests import test_lib


class AttributeContainerIdentifierTest(test_lib.BaseTestCase):
  """Tests for the attribute container identifier."""

  def testCopyToString(self):
    """Tests the CopyToString function."""
    sequence_number = id(self)

    identifier = interface.AttributeContainerIdentifier(
        name='test_container', sequence_number=sequence_number)

    identifier_string = identifier.CopyToString()
    self.assertEqual(
        identifier_string, f'test_container.{sequence_number:d}')


class AttributeContainerTest(test_lib.BaseTestCase):
  """Tests for the attribute container interface."""

  # pylint: disable=protected-access

  def testCopyToDict(self):
    """Tests the CopyToDict function."""
    attribute_container = interface.AttributeContainer()
    attribute_container.attribute_name = 'attribute_name'
    attribute_container.attribute_value = 'attribute_value'

    expected_dict = {
        'attribute_name': 'attribute_name',
        'attribute_value': 'attribute_value'}

    test_dict = attribute_container.CopyToDict()

    self.assertEqual(test_dict, expected_dict)

  def testGetAttributeNames(self):
    """Tests the GetAttributeNames function."""
    attribute_container = interface.AttributeContainer()
    attribute_container._protected_attribute = 'protected'
    attribute_container.attribute_name = 'attribute_name'
    attribute_container.attribute_value = 'attribute_value'

    expected_attribute_names = ['attribute_name', 'attribute_value']

    attribute_names = sorted(attribute_container.GetAttributeNames())

    self.assertEqual(attribute_names, expected_attribute_names)

    attribute_container._SERIALIZABLE_PROTECTED_ATTRIBUTES = [
        '_protected_attribute']

    expected_attribute_names = [
        '_protected_attribute', 'attribute_name', 'attribute_value']

    attribute_names = sorted(attribute_container.GetAttributeNames())

    self.assertEqual(attribute_names, expected_attribute_names)

  def testGetAttributes(self):
    """Tests the GetAttributes function."""
    attribute_container = interface.AttributeContainer()
    attribute_container._protected_attribute = 'protected'
    attribute_container.attribute_name = 'attribute_name'
    attribute_container.attribute_value = 'attribute_value'

    expected_attributes = [
        ('attribute_name', 'attribute_name'),
        ('attribute_value', 'attribute_value')]

    attributes = sorted(attribute_container.GetAttributes())

    self.assertEqual(attributes, expected_attributes)

    attribute_container._SERIALIZABLE_PROTECTED_ATTRIBUTES = [
        '_protected_attribute']

    expected_attributes = [
        ('_protected_attribute', 'protected'),
        ('attribute_name', 'attribute_name'),
        ('attribute_value', 'attribute_value')]

    attributes = sorted(attribute_container.GetAttributes())

    self.assertEqual(attributes, expected_attributes)

  def testGetAttributeValueHash(self):
    """Tests the GetAttributeValuesHash function."""
    attribute_container = interface.AttributeContainer()
    attribute_container._protected_attribute = 'protected'
    attribute_container.attribute_name = 'attribute_name'
    attribute_container.attribute_value = 'attribute_value'

    attribute_values_hash1 = attribute_container.GetAttributeValuesHash()

    attribute_container.attribute_value = 'changes'

    attribute_values_hash2 = attribute_container.GetAttributeValuesHash()

    self.assertNotEqual(attribute_values_hash1, attribute_values_hash2)

    attribute_container.attribute_value = 'attribute_value'

    attribute_container._SERIALIZABLE_PROTECTED_ATTRIBUTES = [
        '_protected_attribute']

    attribute_values_hash2 = attribute_container.GetAttributeValuesHash()

    self.assertNotEqual(attribute_values_hash1, attribute_values_hash2)

  def testGetAttributeValuesString(self):
    """Tests the GetAttributeValuesString function."""
    attribute_container = interface.AttributeContainer()
    attribute_container._protected_attribute = 'protected'
    attribute_container.attribute_name = 'attribute_name'
    attribute_container.attribute_value = 'attribute_value'

    attribute_values_string1 = attribute_container.GetAttributeValuesString()

    attribute_container.attribute_value = 'changes'

    attribute_values_string2 = attribute_container.GetAttributeValuesString()

    self.assertNotEqual(attribute_values_string1, attribute_values_string2)

    attribute_container.attribute_value = 'attribute_value'

    attribute_container._SERIALIZABLE_PROTECTED_ATTRIBUTES = [
        '_protected_attribute']

    attribute_values_string2 = attribute_container.GetAttributeValuesString()

    self.assertNotEqual(attribute_values_string1, attribute_values_string2)

  def testGetIdentifier(self):
    """Tests the GetIdentifier function."""
    attribute_container = interface.AttributeContainer()

    identifier = attribute_container.GetIdentifier()

    self.assertIsNotNone(identifier)

  def testMatchesExpression(self):
    """Tests the MatchesExpression function."""
    attribute_container = interface.AttributeContainer()
    attribute_container.name = 'value'

    result = attribute_container.MatchesExpression('name == "value"')
    self.assertTrue(result)

    result = attribute_container.MatchesExpression('name == "bogus"')
    self.assertFalse(result)

    result = attribute_container.MatchesExpression('bogus')
    self.assertFalse(result)

  def testSetIdentifier(self):
    """Tests the SetIdentifier function."""
    attribute_container = interface.AttributeContainer()

    attribute_container.SetIdentifier(None)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Installation and deployment script."""

import os
import pkg_resources
import sys

try:
  from setuptools import find_packages, setup
except ImportError:
  from distutils.core import find_packages, setup

try:
  from distutils.command.bdist_msi import bdist_msi
except ImportError:
  bdist_msi = None

try:
  from distutils.command.bdist_rpm import bdist_rpm
except ImportError:
  bdist_rpm = None

version_tuple = (sys.version_info[0], sys.version_info[1])
if version_tuple < (3, 7):
  print(f'Unsupported Python version: {sys.version:s}, version 3.7 or higher '
        f'required.')
  sys.exit(1)

# Change PYTHONPATH to include acstore so that we can get the version.
sys.path.insert(0, '.')

import acstore  # pylint: disable=wrong-import-position


if not bdist_msi:
  BdistMSICommand = None
else:
  class BdistMSICommand(bdist_msi):
    """Custom handler for the bdist_msi command."""

    # pylint: disable=invalid-name
    def run(self):
      """Builds an MSI."""
      # Command bdist_msi does not support the library version, neither a date
      # as a version but if we suffix it with .1 everything is fine.
      self.distribution.metadata.version += '.1'

      bdist_msi.run(self)


if not bdist_rpm:
  BdistRPMCommand = None
else:
  class BdistRPMCommand(bdist_rpm):
    """Custom handler for the bdist_rpm command."""

    # pylint: disable=invalid-name
    def _make_spec_file(self):
      """Generates the text of an RPM spec file.

      Returns:
        list[str]: lines of the RPM spec file.
      """
      # Note that bdist_rpm can be an old style class.
      if issubclass(BdistRPMCommand, object):
        spec_file = super(BdistRPMCommand, self)._make_spec_file()
      else:
        spec_file = bdist_rpm._make_spec_file(self)

      python_package = 'python3'

      description = []
      requires = ''
      summary = ''
      in_description = False

      python_spec_file = []
      for line in iter(spec_file):
        if line.startswith('Summary: '):
          summary = line[9:]

        elif line.startswith('BuildRequires: '):
          line = (f'BuildRequires: {python_package:s}-setuptools, '
                  f'{python_package:s}-devel')

        elif line.startswith('Requires: '):
          requires = line[10:]
          continue

        elif line.startswith('%description'):
          in_description = True

        elif line.startswith('python setup.py build'):
          if python_package == 'python3':
            line = '%py3_build'
          else:
            line = '%py2_build'

        elif line.startswith('python setup.py install'):
          if python_package == 'python3':
            line = '%py3_install'
          else:
            line = '%py2_install'

        elif line.startswith('%files'):
          lines = [
              f'%files -n {python_package:s}-%{{name}}',
              '%defattr(644,root,root,755)',
              '%license LICENSE',
              '%doc ACKNOWLEDGEMENTS AUTHORS README']

          lines.extend([
              '%{python3_sitelib}/acstore/*.py',
              '%{python3_sitelib}/acstore/*/*.py',
              '%{python3_sitelib}/acstore*.egg-info/*',
              '',
              '%exclude %{_prefix}/share/doc/*',
              '%exclude %{python3_sitelib}/acstore/__pycache__/*',
              '%exclude %{python3_sitelib}/acstore/*/__pycache__/*'])

          python_spec_file.extend(lines)
          break

        elif line.startswith('%prep'):
          in_description = False

          python_spec_file.append(f'%package -n {python_package:s}-%{{name}}')
          python_summary = f'Python 3 module of {summary:s}'

          if requires:
            python_spec_file.append(f'Requires: {requires:s}')

          python_spec_file.extend([
              f'Summary: {python_summary:s}',
              '',
              f'%description -n {python_package:s}-%{{name}}'])

          python_spec_file.extend(description)

        elif in_description:
          # Ignore leading white lines in the description.
          if not description and not line:
            continue

          description.append(line)

        python_spec_file.append(line)

      return python_spec_file


def parse_requirements_from_file(path):
  """Parses requirements from a requirements file.

  Args:
    path (str): path to the requirements file.

  Returns:
    list[str]: name and optional version information of the required packages.
  """
  requirements = []
  if os.path.isfile(path):
    with open(path, 'r') as file_object:
      file_contents = file_object.read()

    for requirement in pkg_resources.parse_requirements(file_contents):
      try:
        name = str(requirement.req)
      except AttributeError:
        name = str(requirement)

      if not name.startswith('pip '):
        requirements.append(name)

  return requirements


acstore_description = (
    'Attribute Container Storage (ACStore).')

acstore_long_description = (
    'ACStore, or Attribute Container Storage, provides a stand-alone '
    'implementation to read and write attribute container storage files.')

command_classes = {}
if BdistMSICommand:
  command_classes['bdist_msi'] = BdistMSICommand
if BdistRPMCommand:
  command_classes['bdist_rpm'] = BdistRPMCommand

setup(
    name='acstore',
    version=acstore.__version__,
    description=acstore_description,
    long_description=acstore_long_description,
    long_description_content_type='text/plain',
    license='Apache License, Version 2.0',
    url='https://github.com/log2timeline/acstore',
    maintainer='Log2Timeline maintainers',
    maintainer_email='log2timeline-maintainers@googlegroups.com',
    cmdclass=command_classes,
    classifiers=[
        '',
        'Environment :: Console',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    packages=find_packages('.', exclude=[
        'docs', 'tests', 'tests.*', 'utils']),
    package_dir={
        'acstore': 'acstore'
    },
    data_files=[
        ('share/doc/acstore', [
            'ACKNOWLEDGEMENTS', 'AUTHORS', 'LICENSE', 'README']),
    ],
    install_requires=parse_requirements_from_file('requirements.txt'),
    tests_require=parse_requirements_from_file('test_requirements.txt'),
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the attribute container manager."""

import unittest

from acstore.containers import manager

from tests import test_lib as shared_test_lib


class AttributeContainersManagerTest(shared_test_lib.BaseTestCase):
  """Tests for the attribute container manager."""

  # pylint: disable=protected-access

  def testCreateAttributeContainer(self):
    """Tests the CreateAttributeContainer function."""
    manager.AttributeContainersManager.RegisterAttributeContainer(
        shared_test_lib.TestAttributeContainer)

    try:
      attribute_container = (
          manager.AttributeContainersManager.CreateAttributeContainer(
              'test_container'))
      self.assertIsNotNone(attribute_container)

      with self.assertRaises(ValueError):
        manager.AttributeContainersManager.CreateAttributeContainer('bogus')

    finally:
      manager.AttributeContainersManager.DeregisterAttributeContainer(
          shared_test_lib.TestAttributeContainer)

  def testGetContainerTypes(self):
    """Tests the GetContainerTypes function."""
    manager.AttributeContainersManager.RegisterAttributeContainer(
        shared_test_lib.TestAttributeContainer)

    try:
      container_types = manager.AttributeContainersManager.GetContainerTypes()
      self.assertIn('test_container', container_types)

    finally:
      manager.AttributeContainersManager.DeregisterAttributeContainer(
          shared_test_lib.TestAttributeContainer)

  def testGetSchema(self):
    """Tests the GetSchema function."""
    manager.AttributeContainersManager.RegisterAttributeContainer(
        shared_test_lib.TestAttributeContainer)

    try:
      schema = manager.AttributeContainersManager.GetSchema('test_container')
      self.assertIsNotNone(schema)
      self.assertEqual(schema, shared_test_lib.TestAttributeContainer.SCHEMA)

      with self.assertRaises(ValueError):
        manager.AttributeContainersManager.GetSchema('bogus')

    finally:
      manager.AttributeContainersManager.DeregisterAttributeContainer(
          shared_test_lib.TestAttributeContainer)

  def testAttributeContainerRegistration(self):
    """Tests the Register and DeregisterAttributeContainer functions."""
    number_of_classes = len(
        manager.AttributeContainersManager._attribute_container_classes)

    manager.AttributeContainersManager.RegisterAttributeContainer(
        shared_test_lib.TestAttributeContainer)

    try:
      self.assertEqual(
          len(manager.AttributeContainersManager._attribute_container_classes),
          number_of_classes + 1)

      with self.assertRaises(KeyError):
        manager.AttributeContainersManager.RegisterAttributeContainer(
            shared_test_lib.TestAttributeContainer)

    finally:
      manager.AttributeContainersManager.DeregisterAttributeContainer(
          shared_test_lib.TestAttributeContainer)

    self.assertEqual(
        len(manager.AttributeContainersManager._attribute_container_classes),
        number_of_classes)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the attribute container store interface."""

import unittest

from acstore import interface
from acstore.containers import manager

from tests import test_lib


class AttributeContainerStoreTest(test_lib.BaseTestCase):
  """Tests for the attribute container store interface."""

  # pylint: disable=protected-access

  def testGetAttributeContainerNextSequenceNumber(self):
    """Tests the _GetAttributeContainerNextSequenceNumber function."""
    attribute_container = test_lib.TestAttributeContainer()

    test_store = interface.AttributeContainerStore()

    sequence_number = test_store._GetAttributeContainerNextSequenceNumber(
        attribute_container.CONTAINER_TYPE)
    self.assertEqual(sequence_number, 1)

    sequence_number = test_store._GetAttributeContainerNextSequenceNumber(
        attribute_container.CONTAINER_TYPE)
    self.assertEqual(sequence_number, 2)

  def testGetAttributeContainerSchema(self):
    """Tests the _GetAttributeContainerSchema function."""
    attribute_container = test_lib.TestAttributeContainer()

    test_store = interface.AttributeContainerStore()

    schema = test_store._GetAttributeContainerSchema(
        attribute_container.CONTAINER_TYPE)
    self.assertEqual(schema, {})

    manager.AttributeContainersManager.RegisterAttributeContainer(
        test_lib.TestAttributeContainer)

    try:
      schema = test_store._GetAttributeContainerSchema(
          attribute_container.CONTAINER_TYPE)
      self.assertEqual(schema, test_lib.TestAttributeContainer.SCHEMA)

    finally:
      manager.AttributeContainersManager.DeregisterAttributeContainer(
          test_lib.TestAttributeContainer)

  # TODO: add tests for _SetAttributeContainerNextSequenceNumber

  def testSetStorageProfiler(self):
    """Tests the SetStorageProfiler function."""
    test_store = interface.AttributeContainerStore()
    test_store.SetStorageProfiler(None)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the profiler classes."""

import time
import unittest

from acstore import profilers

from tests import test_lib


class CPUTimeMeasurementTest(test_lib.BaseTestCase):
  """Tests for the CPU time measurement."""

  def testSampleStartStop(self):
    """Tests the SampleStart and SampleStop functions."""
    cpu_measurement = profilers.CPUTimeMeasurement()
    cpu_measurement.SampleStart()
    cpu_measurement.SampleStop()


class StorageProfilerTest(test_lib.BaseTestCase):
  """Tests for the storage profiler."""

  # pylint: disable=protected-access

  def testIsSupported(self):
    """Tests the IsSupported function."""
    self.assertTrue(profilers.StorageProfiler.IsSupported())

  def testStartStop(self):
    """Tests the Start and Stop functions."""
    with test_lib.TempDirectory() as temp_directory:
      test_profiler = profilers.StorageProfiler(
          'test', temp_directory)

      test_profiler._FILENAME_PREFIX = 'test'
      test_profiler._FILE_HEADER = 'test'

      test_profiler.Start()

      test_profiler.Stop()

  def testSample(self):
    """Tests the Sample function."""
    with test_lib.TempDirectory() as temp_directory:
      test_profiler = profilers.StorageProfiler(
          'test', temp_directory)

      test_profiler.Start()

      for _ in range(5):
        test_profiler.StartTiming('test_profile')
        time.sleep(0.01)
        test_profiler.StopTiming('test_profile')
        test_profiler.Sample('test_profile', 'read', 'test', 1024, 128)

      test_profiler.Stop()


if __name__ == '__main__':
  unittest.main()
